type
    THeader = record
        password, version, code, bpp : byte;
        x1, y1, x2, y2, horiz_res, vert_res : integer;
        palette : array[0..pred(16*3)] of byte;
        vmode, nplanes : byte;
        bpl, palinfo, shres, svres : integer;
        dumy : array[0..53] of char;
    end;
    PBuffer = ^TBuffer;
    TBuffer = array[0..pred(64000)] of byte;
const
    filename : string = 'test.pcx';
    VESA_MODE         = $112;
    BUFFER_SIZE       = 10000;
    CURRENT_BANK      : word = 0;
    current_ptr       : word = 0;
    end_ptr           : word = 0;
var
    fp                     : file;
    data_buffer            : PBuffer;
    write_bank, shift_bank : word;
    header                 : THeader;


function readData : byte;
begin
    if current_ptr >= end_ptr then begin
       BlockRead(fp,data_buffer^,BUFFER_SIZE,end_ptr);
       current_ptr := 0;
    end;
    readData := data_buffer^[current_ptr];
    inc(current_ptr);
end;

procedure PutPixel(x, y : longint; b, g, r : byte);
var
    color : longint;
begin
asm
    db   66h;    mov  cx, word(x)
    db   66h;    mov  bx, word(y)
    db   66h;    cmp  cx, 0
                 jb   @@END
    db   66h;    cmp  bx, 0
                 jb   @@END
    db   66h;    cmp  cx, 640;          dw   0000h
                 ja   @@END
    db   66h;    cmp  bx, 480;          dw   0000h
                 ja   @@END
                 mov  ax, 0A000h
                 mov  es, ax
                 lea  di, color
                 mov  al, r
                 mov  ss:[di], al
                 mov  al, g
                 mov  ss:[di+1], al
                 mov  al, b
                 mov  ss:[di+2], al
    db   66h;    mov  ax, 1920;         dw   0000h
    db   66h;    mul  bx
    db   66h;    mov  dx, cx
    db   66h;    shl  cx, 1
    db   66h;    add  cx, dx
    db   66h;    add  ax, cx
    db   66h;    mov  bx, ax
    db   66h;    shr  bx, 16
                 mov  dx, CURRENT_BANK
    db   66h;    xchg ax, bx

                 { bx : offset, ax : bank, cx : loop, dx : CURRECT BANK }
                 xor  cx, cx
                 cmp  ax, dx
                 je   @@LOOP

@@VERIFY_BANK:
                 mov  CURRENT_BANK, ax
                 push ax; push bx; push cx; push dx
                 mov  ax, 4F05h
                 mov  bx, write_bank
                 mov  dx, CURRENT_BANK
                 mov  cx, shift_bank
                 shl  dx, cl
                 int  10h
                 pop dx; pop cx; pop bx; pop ax

@@LOOP:
                 mov  ch, ss:[di]
                 mov  es:[bx], ch
                 xor  ch, ch
                 inc  di

                 inc  cx
                 cmp  cx, 3
                 je   @@END
                 inc  bx
                 cmp  bx, 0
                 jne  @@LOOP
                 inc  ax
                 jmp  @@VERIFY_BANK

@@END:
    end;
end;

procedure decodePCX24(var header : THeader);
var
    i, num : integer;
    data : byte;
    x, y, width, height, byte_p_line : integer;
    buffer : PBuffer;
begin
    getMem(buffer,header.bpl * header.nplanes);

    width  := succ(header.x2 - header.x1);
    height := succ(header.y2 - header.y1);
    byte_p_line := header.bpl*header.nplanes;

    for y := 0 to pred(height) do begin
        i := 0;
        while (i < byte_p_line) do begin
            data := readData;
            if data >= $C0 then begin
                num := data and $3F;
                data := readData;
                while num > 0 do begin
                    buffer^[i] := data;
                    inc(i);
                    dec(num);
                end;
            end
            else begin
                buffer^[i] := data;
                inc(i);
            end;
        end;
        for x := 0 to pred(width) do begin
            PutPixel(x+(640-width) div 2,y+(480-height) div 2,
                buffer^[x],buffer^[x+header.bpl],buffer^[x+header.bpl*2]);
        end;
    end;
end;

procedure initGraph;
type
   TVESAInfo = record
      Mode_Attribute : word;
      Window_A_Attribute : byte;
      Window_B_Attribute : byte;
      Window_Granularity : word;
      gabage : array[1..250] of byte;
   end;
var
   VESAinfo  : TVESAInfo;
   pVESAinfo : ^TVESAInfo;
   result    : word;
begin

   pVESAinfo := @VESAinfo;
   asm
      mov  ax, 4F01h
      mov  cx, VESA_MODE
      les  di, pVESAinfo
      int  10h
      mov  result, ax
   end;

   if result <> $004F then begin
      write(^G);
      writeLn('VESA DRIVER not supported.');
      halt;
   end;

   if (pVESAinfo^.Window_A_Attribute and 5) = 5 then write_bank := 0;
   if (pVESAinfo^.Window_B_Attribute and 5) = 5 then write_bank := 1;

   for shift_bank := 0 to 15 do begin
      if ($40 shr shift_bank) = pVESAinfo^.Window_Granularity then break;
   end;

   asm
      mov  ax, 4F02h
      mov  bx, VESA_MODE
      int  10h
      mov  byte ptr result, ah
   end;

   if byte(result) = 1 then begin
      write(^G);
      writeLn('This mode not supported.');
      halt;
   end;

end;

procedure closeGraph; assembler;
asm
      mov  ax, 4F02h
      mov  bx, 03h
      int  10h
end;

begin
    initGraph;

    getMem(data_buffer,BUFFER_SIZE);

    assign(fp,filename);
    {$I-}
    reset(fp,1);
    {$I+}
    if IOResult <> 0 then begin
       closeGraph;
       writeLn('File not found.');
       halt;
    end;

    BlockRead(fp,header,sizeof(header));

    if (header.password <> 10) or (header.bpp * header.nplanes <> 8*3) then exit;
    decodePCX24(header);

    close(fp);
    freeMem(data_buffer,BUFFER_SIZE);

    readln;

    closeGraph;
end.
