type
    THeader = record
        password, version, code, bpp : byte;
        x1, y1, x2, y2, horiz_res, vert_res : integer;
        palette : array[0..pred(16*3)] of byte;
        vmode, nplanes : byte;
        bpl, palinfo, shres, svres : integer;
        dumy : array[0..53] of char;
    end;
    PBuffer = ^TBuffer;
    TBuffer = array[0..pred(64000)] of byte;
const
    filename : string = 'test.pcx';
    VESA_MODE         = $103;
    MAX_X_AXIS        = 800;
    MAX_Y_AXIS        = 600;
    BUFFER_SIZE       = 10000;
    CURRENT_BANK      : word = 0;
var
    fp                     : file;
    data_buffer            : ^TBuffer;
    current_ptr, end_ptr   : word;
    write_bank, shift_bank : word;
    header                 : THeader;


function readData : byte;
begin
    if current_ptr >= end_ptr then begin
       BlockRead(fp,data_buffer^,BUFFER_SIZE,end_ptr);
       current_ptr := 0;
    end;
    readData := data_buffer^[current_ptr];
    inc(current_ptr);
end;

procedure PutPixel(x, y : longint; b, g, r : byte); assembler;
asm
    db   66h;    mov  cx, word(x)
    db   66h;    mov  bx, word(y)
    db   66h;    cmp  cx, 0
                 jb   @@END
    db   66h;    cmp  bx, 0
                 jb   @@END
    db   66h;    cmp  cx, MAX_X_AXIS;   dw   0000h
                 ja   @@END
    db   66h;    cmp  bx, MAX_Y_AXIS;   dw   0000h
                 ja   @@END

                 mov  ax, 0A000h;
                 mov  es, ax
    db   66h;    mov  ax, MAX_X_AXIS;   dw   0000h
    db   66h;    mul  bx
    db   66h;    add  ax, cx
    db   66h;    mov  dx, ax
    db   66h;    shr  dx, 16
                 mov  bx, CURRENT_BANK
                 mov  di, ax

                 cmp  bx, dx
                 je   @@SAME_BANK
                 mov  CURRENT_BANK, dx
                 mov  ax, 4F05h
                 mov  bx, write_bank
                 mov  cx, shift_bank
                 shl  dx, cl
                 int  10h

@@SAME_BANK:
                 mov  bx, 5
                 mov  al, b
                 xor  ah, ah
                 mul  bx
                 shr  ax, 8
                 mov  cx, ax
                 mov  al, g
                 xor  ah, ah
                 mul  bx
                 shr  ax, 8
                 inc  bx
                 mul  bx
                 add  cx, ax
                 dec  bx
                 mov  al, r
                 xor  ah, ah
                 mul  bx
                 shr  ax, 8
                 mov  bx, 36
                 mul  bx
                 add  ax, cx

                 mov  es:[di], al
@@END:
end;

procedure _PutPixel(x, y : longint; b, g, r : byte);
var
    temp : longint;
    i, bank, offset : word;
    color : byte;
begin
    if (x < 0) or (y < 0) or (x > MAX_X_AXIS) or (y > MAX_Y_AXIS) then exit;

    temp := longint(y) * MAX_X_AXIS + x;
    bank := temp shr 16;
    offset := word(temp);
    if CURRENT_BANK <> bank then begin
        CURRENT_BANK := bank;
        asm
                mov  ax, 4F05h
                mov  bx, write_bank
                mov  dx, CURRENT_BANK
                mov  cx, shift_bank
                shl  dx, cl
                int  10h
        end;
    end;
    color := (r*5 div 255)*36 + (g*5 div 255)*6 + (b*5 div 255);
    Mem[$A000:offset] := color;
end;

procedure decodePCX16bit(var header : THeader);
var
    i, num : integer;
    data : byte;
    x, y, width, height, byte_p_line : integer;
    buffer : PBuffer;
begin
    getMem(buffer,header.bpl * header.nplanes);

    width  := succ(header.x2 - header.x1);
    height := succ(header.y2 - header.y1);
    byte_p_line := header.bpl*header.nplanes;

    for y := 0 to pred(height) do begin
        i := 0;
        while (i < byte_p_line) do begin
            data := readData;
            if data >= $C0 then begin
                num := data and $3F;
                data := readData;
                while num > 0 do begin
                    buffer^[i] := data;
                    inc(i);
                    dec(num);
                end;
            end
            else begin
                buffer^[i] := data;
                inc(i);
            end;
        end;
        for x := 0 to pred(width) do begin
            PutPixel(x+(MAX_X_AXIS-width) div 2,y+(MAX_Y_AXIS-height) div 2,
                buffer^[x],buffer^[x+header.bpl],buffer^[x+header.bpl*2]);
        end;
    end;

end;

procedure initGraph;
type
   TVESAInfo = record
      Mode_Attribute : word;
      Window_A_Attribute : byte;
      Window_B_Attribute : byte;
      Window_Granularity : word;
      gabage : array[1..250] of byte;
   end;
var
   VESAinfo  : TVESAInfo;
   pVESAinfo : ^TVESAInfo;
   result, i : word;
begin

   pVESAinfo := @VESAinfo;
   asm
      mov  ax, 4F01h
      mov  cx, VESA_MODE
      les  di, pVESAinfo
      int  10h
      mov  result, ax
   end;

   if result <> $004F then writeLn('VESA DRIVER not supported.');

   if (pVESAinfo^.Window_A_Attribute and 5) = 5 then write_bank := 0;
   if (pVESAinfo^.Window_B_Attribute and 5) = 5 then write_bank := 1;

   for shift_bank := 0 to 15 do begin
      if ($40 shr shift_bank) = pVESAinfo^.Window_Granularity then break;
   end;

   asm
      mov  ax, 4F02h
      mov  bx, VESA_MODE
      int  10h
      mov  byte ptr result, ah
   end;
   if byte(result) = 1 then begin
      write(^G);
      writeLn('This mode not supported.');
      halt;
   end;

   port[$3c8] := 0;
   for i := 0 to 6*6*6 do begin
      port[$3c9] := 63 * (i mod 6) div 6;
      port[$3c9] := 63 * ((i div 6) mod 6) div 6;
      port[$3c9] := 63 * (i div 36) div 6;
   end;

end;

procedure closeGraph; assembler;
asm
    mov  ax, 4F02h
    mov  bx, 03h
    int  10h
end;

begin
    initGraph;

    current_ptr := 0; end_ptr := 0;
    getMem(data_buffer,BUFFER_SIZE);

    assign(fp,filename);
    reset(fp,1);
    BlockRead(fp,header,sizeof(header));

    if (header.password <> 10) or (header.bpp * header.nplanes <> 8*3) then exit;
    decodePCX16bit(header);

    close(fp);
    freeMem(data_buffer,BUFFER_SIZE);

    readln;

    closeGraph;
end.
